/*

 DDS Synthesizer with Analog Devices AD9850 chip
 
 Juha Niinikoski OH2NLT 28.03.2004

 Note if you use 16F877A chip add -NO_STRING_PACK option for correct putst() operation.

 ! Beware faulty 16F877A (shit) parts. Two weeks work for nothing to find this. ! 22.01.2004
                               Read Microchip Errata papers.

 You can use either 16F877A or 16F877 chips. With 16F877A you get better back light control

 EEPROM routines added 03.04.2004
 Fixed point frequency calculations, Osc. calibration & some IRQ routine cleaning 05.04.2004
 
 SW3 Menu down added, Display Freq = Freq/4 19.4.2004

*/

//*********************
//include header files
//*********************

#include	<pic.h>
#include	"dds.h"


//	__CONFIG(UNPROTECT|BODEN|FOSC1|BKBUG|WRT|CPD);		// CPU config bits v7.xx
	__CONFIG(UNPROTECT&WDTDIS&BOREN&HS&DEBUGDIS&PWRTEN&DUNPROT&PWRTEN&LVPDIS); // config bits for v8.01


//*********************
// Global variables 
//*********************

//const char   banner[] = {"    OH2NLT"};	// banner
const char  banner[] = {"  OH2NLT/OH7SV"};	// banner
const char _version[] = {" DDS 29.04.2004"};	// system version info


#define DEF_MODE 1				// default display mode
#define DEF_INT 7				// default display intensity

char d_mode;					// display mode
bit redraw;					// something changed redraw display
bit service_mode;				// service mode selected

bank1 char dac;					// Voltage reference module (DAC) value
bank1 int adc_ch0;				// adc raw values
bank1 int adc_ch1;

bit pw_ready;					// pulse wheel data available
long f_step;					// wheel +/- step

const long f_steps[3] = {1, 1000, 1000000};	// increment / decrement steps
bank1 char step_idx;				// step table index

	struct dds_regs
	{
	unsigned long dds_freq;
	unsigned char dds_config;
	};

	union					// AD9850 frequency & config data, index 0 = LSB
	{
	struct dds_regs dds_regs;
	unsigned char ad9850[5];
	} dds;


#define MIN_DDS 0				// DDS limits
//#define MAX_DDS 1374389535			// 40 MHz upper limit
#define MAX_DDS 1717986918			// 50 MHz upper limit

#define DEFAULT_FREQ 508386689			// "factory default" 4 * 3.699 MHz

#define CLKIN 125000000				// oscilator 125 MHz


/* Calibration data */

	struct caldata
	{
	unsigned long osc_freq;			// oscilator frequency with +/- correction
	unsigned long start_freq;		// startup frequency stting
	unsigned char start_step;		// startup frequency step
	unsigned char start_dmode;		// startup display mode
	unsigned char start_dac;		// startup display back ligth setting
	unsigned char spare[21];
	unsigned char cal_csum;			// calibration EEPROM data block checksum
	};

	union
	{
	struct caldata caldata;			// map calibration data to EEPROM registers
	unsigned char eeprom[32];		// eeprom[31] reserved for checksum
	}bank1 eedata;


//*********************
// Include files 
//*********************

#include	"timers.c"
#include	"lcd8.c"
#include	"term.c"
#include	"adc.c"
#include 	"eeprom.c"
#include	"ad9850.c"




//****************************************
// Clumsy but working 32*32 multiply JNi
// elemantary school scheme but with
// base 256 numbers.
//****************************************

// bank2 variables, problems with RAM space

	union					// 32-bit term 1
	{
	unsigned long m1l;
	unsigned char m1b[4];
	}bank2 m1;

	union					// 32-bit term 2
	{
	unsigned long m2l;
	unsigned char m2b[4];
	}bank2 m2;

	struct prodlng
	{
	unsigned long prodlo;
	unsigned long prodhi;
	};

	union					// 64-bit product register
	{
	struct prodlng prodlng;
	unsigned char pb[8];
	}bank2 prod;


	union					// digit sum register
	{
	unsigned int dsi;
	unsigned char dsc[2];
	}bank2 ds;

	union					// product sum register
	{
	unsigned int psi;
	unsigned char psc[2];
	}bank2 ps;


#if 0
void pdebug(char d1, char d2)		// debug printouts
	{
	char x;
	sio_putst("\n\n d1, d2 = ");
	sio_putchhex(d1);
	sio_putch(' ');
	sio_putchhex(d2);

	sio_putst("\nm1 ");
	for(x=0; x<4; x++)
		{
		sio_putchhex(m1.m1b[x]);
		sio_putch(' ');
		}

	sio_putst("\nm2 ");
	for(x=0; x<4; x++)
		{
		sio_putchhex(m2.m2b[x]);
		sio_putch(' ');
		}

	sio_putst("\npr ");
	for(x=0; x<8; x++)
		{
		sio_putchhex(prod.pb[x]);
		sio_putch(' ');
		}
	}
#endif

void add_prod(char idx, unsigned char d)			// 8-bit to product array idx position
		{
		ps.psi = prod.pb[idx] + d;			// sum digit
		prod.pb[idx] = ps.psc[0];
		while((ps.psc[1] != 0) || idx < 8)		// continue adding if carry
			{
			idx++;
			ps.psi = prod.pb[idx] + ps.psc[1];
			prod.pb[idx] = ps.psc[0];
			}	
		}


void clear_prod(void)						// clear product array
	{
	char x;
	for(x=0; x<8; x++)
		{
		prod.pb[x] = 0;
		}
	}	


void vlmul(void)						// 32*32 multiply. prod = m1*m2
	{
	char d1, d2;
	clear_prod();

	for(d2=0; d2<4; d2++)
		{
		for(d1=0; d1<4; d1++)
			{
			ds.dsi = m1.m1b[d1] * m2.m2b[d2];	// multiply digit
			add_prod((d1+d2), ds.dsc[0]);		// add to product
			if(ds.dsc[1] != 0)
				add_prod((d2+d1+1), ds.dsc[1]);	// add carry number
			
			}
		}
	}




//***************************
// Helper & I/O functions
//***************************


/* SW2 functions */

void freq_step_adj(void)
	{
	if(SW2 == 0)				// addjust freq step
		{
		step_idx++;
		if(step_idx >= 3)
			step_idx = 0;
		f_step = f_steps[step_idx];

		delay(250);			// button debounce / repeat
		delay(150);
		redraw = 1;			// force update
		}
	}



//***************************
// DDS Main
//***************************

void main( void )
{
int temp;
char dly;

/* Init IO Ports */

// Set up PORTA
	TRISA = TRISA_INIT;
	PORTA = PORTA_INIT;
	ADCON0 = ADCON0_INIT;
	ADCON1 = ADCON1_INIT;

// Set up PORTB
	TRISB = TRISB_INIT;
	PORTB = PORTB_INIT;

// Set up PORTC
	TRISC = TRISC_INIT;
	PORTC = PORTC_INIT;

// Set up PORTD
	TRISD = TRISD_INIT;
	PORTD = PORTD_INIT;

// Set up PORTE
	TRISE = TRISE_INIT;
	PORTE = PORTE_INIT;


// init LCD controller

	initlcd();			// init LCD system
	set_chgen();			// set bar graph fonts

	
// init serial port

	serial_setup();			


// print hello messages

	putst(banner);			// hello to LCD
	set_cur_lcd(LINE2);
	putst(_version);

	sio_putst("\n");		// hello to serial port
	sio_putst(banner);
	sio_putch(' ');
	sio_putst(_version);
	sio_putst("\n  Prototype running\n");

	for(dly = 0; dly < 16; dly++)	// show Hello message for a while
		{
		delay( 250 );
		}

// check if service mode start. SW1 pressed while start up

	if(SW1 == 0)
		{
		service_mode = 1;	// set service mode
		clear_lcd();
		putst("Service Mode");
		while(SW1 == 0)
			;
		}

// init operating variables from EEPROM

	read_eeprom_data();		// read defaults from EEPROM
	SIDEBAND = LSB;


// init irq system & timers

	Init_timers();			// init timers & irq system


// init DDS

	dds_init();			// init AD9850
	dds_load();			// do first load
	dds_load();	// test, start up problem

// start main process loop

	clear_lcd();			// start with clear screen

for (;;)
	{

#define MAX_NORMAL 4			// highest normal display mode number
#define MAX_DEBUG 7			// highest debug display mode number


// do switches

	if(SW1 == 0 || SW3 == 0)	// select display mode
		{
		if(SW1 == 0)		// display mode up
			{
			delay(50);
			d_mode++;
			if(service_mode == 0)	// limit normal mode selections
				if(d_mode >= (MAX_NORMAL+1))
					d_mode = 1;
			else
				if(d_mode >= (MAX_DEBUG+1))	
					d_mode = 1;
			}

		if(SW3 == 0)		// display mode down
			{
			delay(50);
			d_mode--;
			if(d_mode < 1)
				{
				if(service_mode == 0)	// limit normal mode selections
					d_mode = MAX_NORMAL;
				else
					d_mode = MAX_DEBUG;
				}
			}

		clear_lcd();
		putst("Mode ");
		putchhex(d_mode);
		delay(250);
		while(SW1 == 0 || SW3 == 0);		// show mode as long as pressed
			;
		clear_lcd();
		redraw = 1;
		if(d_mode == 7)		// Calibration mode needs preparations
			{
			set_pw_data(eedata.caldata.osc_freq); // osc to pulse wheel
			step_idx = 0;
			f_step = f_steps[step_idx];	// 1 Hz step
			}
		delay(100);		// button debounce / repeat
		}

// do DDS load. Measured execution time 185us

	if(pw_ready && (d_mode != 7))			// read pulse wheel if new data available
		{
		pw_ready = 0;
		redraw = 1;				// force redraw	
 
		dds.dds_regs.dds_freq = get_pw_data();
//		dds.dds_regs.dds_freq = pw_data;    // test

		dds_load();
		}


// measure ADC values for future calculations

adc_ch0 = convert_adc(0);
adc_ch1 = convert_adc(1);



// do selected display page & functions

	switch (d_mode)
		{

		case 1:					// Main Frequency display
	LOOP_TEST = 1;	// test, measure loop performance. Measured 6ms.
			redraw = 0;
			set_cur_lcd(LINE1);
			draw_s_meter((char)(adc_ch0/12)); // draw S-meter bar

			set_cur_lcd(LINE2);
			m1.m1l = dds.dds_regs.dds_freq;	  // calculate display frequency
			m2.m2l = eedata.caldata.osc_freq; // f = (pw_data*osc_freq)/2^32
			vlmul();			  // do JNi 32*32 multiplication
			putlongdec((prod.prodlng.prodhi >> 2), 1);  // use hi-part = divided 2^32 / 4
			putst(" Hz ");

			if(SIDEBAND == LSB)			// show sideband
				putch('L');
			else
				putch('U');

			put1hex(step_idx);		// show DDS tuning step
	LOOP_TEST = 0;	

			freq_step_adj();		// addjust freq step	
			break;


		case 2:					// S-meter test display
			set_cur_lcd(LINE1);
			putst("AD ");
			putintdec(adc_ch0, 0);	// test
			putintdec(adc_ch1, 0);

			if(SW2 == 0)
				{
				SIDEBAND = !SIDEBAND;	// change sideband
				delay(250);
				}

			if(SIDEBAND == LSB)		// test sideband selector
				putst("  LSB");
			else
				putst("  USB");


			set_cur_lcd(LINE2);
			temp = adc_ch0;
			temp = temp/12;			// scale for bar, 0-5V input range
			draw_bar((char)(adc_ch0/12), 0);	// draw meter bar

			break;


		case 3:					// store in use operating values to EEPROM 
			set_cur_lcd(LINE1);		// need more thinking
			putst("Store Defaults");
			set_cur_lcd(LINE2);				
			putst("SW2 = Write");
			if(SW2 == 0)
				{
				d_mode = DEF_MODE;	// do this while better ideas appears
				write_eeprom_data();
				clear_lcd();
				putst("Write EEPROM");
				set_cur_lcd(LINE2);				
				putst("Ok.");
				delay(250);
				delay(250);
				delay(250);
				while(SW2 == 0)
					;
				delay(250);
				}
			break;


		case 4:					// Back light DAC or I/O bit
			set_cur_lcd(LINE1);			
			putst("Back Light Set");
			set_cur_lcd(LINE2);		// print setting
			putst("   ");
#ifdef __A_TYPE
			putintdec(dac, 0);		// if 877A DAC
#else
			if(BL)
				putst("On  ");		// if 877 on/off control
			else
				putst("Off ");
#endif

			if(SW2 == 0)			// adjust back light with SW2
				{
#ifdef __A_TYPE
				dac++;			// addlust DAC if we have 877A chip
				if(dac > 16)
					dac = 0;
				set_bl(dac);
#else
				BL = !BL;		// else on / off toggle
#endif
			set_cur_lcd(LINE2);		// reprint after setting
			putst("   ");
#ifdef __A_TYPE
			putintdec(dac, 0);		// if 877A DAC
#else
			if(BL)
				putst("On  ");		// if 877 on/off control
			else
				putst("Off ");
#endif
				delay(250);		// button debounce / repeat
				}

			break;


/* service mode displays & commands */

		case 5:						// DDS raw set value display

			set_cur_lcd(LINE1);
			putchhex(step_idx);
			putch(' ');

//			putchhex(pw_sample);	// pulse wheel test
//			putch(' ');
//			putchhex(PORTB & 7);	// pulse wheel test
//			putch(' ');

			putlongdec(f_step, 0);

			set_cur_lcd(LINE2);

//			putchhex(pw_state);	// pulse wheel test
//			putst(" ");
//			if(INTEDG == 1)		// test
//				putst("1");
//			else
//				putst("0");
//			putst(" ");				

			putst("N  ");
			putlongdec(dds.dds_regs.dds_freq, 0); // display DDS set value

			freq_step_adj();			// addjust freq step
			break;


		case 6:						// load JNi "factory defaults" 
			set_cur_lcd(LINE1);
			putst("Factory Setup ?");
			set_cur_lcd(LINE2);				
			putst("SW2 = Load");
			if(SW2 == 0)
				{
				d_mode = DEF_MODE;		// do this while better ideas appears
				redraw = 1;
				load_defaults();
				write_eeprom_data();
				clear_lcd();				
				putst("Ok.");
				delay(250);
				delay(250);
				delay(250);
				while(SW2 == 0)
					;
				delay(250);
				}
			break;

		case 7:						// Calibrate 125MHz Oscilator 
			set_cur_lcd(LINE1);
			putst("Calibrate SW2");
			set_cur_lcd(LINE2);
			eedata.caldata.osc_freq = get_pw_data(); // addjust oscilator frequency setting	
			putlongdec(eedata.caldata.osc_freq, 0);
			putst(" Hz");

			if(SW2 == 0)
				{
				write_eeprom_data();		// save everything
				d_mode = DEF_MODE;		// return to normal mode
				set_pw_data(dds.dds_regs.dds_freq); // restore display freq
				redraw = 1;
				clear_lcd();				
				putst("Ok.");
				delay(250);
				delay(250);
				delay(250);
				while(SW2 == 0)
					;
				delay(250);
				}
			break;


		default: case 0:
			d_mode = 1;			// push to available displays range
			break;
		}

/* some serial port routines & tests */

	if(keypressed() == 1)
		{

		switch (sio_getch())
			{

			case 'd':
			dump_eeprom();
			break;

			case 'c':
			sio_putst("\nChecksum: ");
			sio_putchhex(calc_csum());
			break;

			case 'w':
			write_eeprom_data();
			break;

			case 'r':
			read_eeprom_data();
			break;

			}		
		}
	}
}

